#include "dynamics.h"

#include "slideprimitive.h"
#include "slideworld.h"
#include <tcl.h>

VOID DynEval(FLOAT *pfValue, CHAR **ppcDynExpr, UINT_32 uiComponents)
{
  UINT_32 i;
  INT_32 iCode;

  ASSERT( pfValue != NULL );
  ASSERT( ppcDynExpr != NULL );

  for ( i = 0; i < uiComponents; i++ )
    {
      if ( ppcDynExpr[i] != NULL )
	{
	  iCode = Tcl_Eval(CSLIDEWorld::GetInterp(), ppcDynExpr[i]);
	  if ( iCode != TCL_OK )
	    {
	      fprintf(CSLIDEPrimitive::GetErrorFile(), 
		      "Error: Tcl returned error code \"%s\" for the block:\n%s\n",
		      CSLIDEWorld::GetInterp()->result, ppcDynExpr[i]);
	      exit(1);
	    }
	  else
	    {
	      pfValue[i] = (FLOAT)atof(CSLIDEWorld::GetInterp()->result);
	    }
	}      
    }
}

VOID DynEvalInt(INT_32 *piValue, CHAR **ppcDynExpr, UINT_32 uiComponents)
{
  UINT_32 i;
  INT_32 iCode;

  ASSERT( piValue != NULL );
  ASSERT( ppcDynExpr != NULL );

  for ( i = 0; i < uiComponents; i++ )
    {
      if ( ppcDynExpr[i] != NULL )
	{
	  iCode = Tcl_Eval(CSLIDEWorld::GetInterp(), ppcDynExpr[i]);
	  if ( iCode != TCL_OK )
	    {
	      fprintf(CSLIDEPrimitive::GetErrorFile(), 
		      "Error: Tcl returned error code \"%s\" for the block:\n%s",
		      CSLIDEWorld::GetInterp()->result, ppcDynExpr[i]);
	      exit(1);
	    }
	  else
	    {
	      piValue[i] = atol(CSLIDEWorld::GetInterp()->result);
	    }
	}      
    }
}

//////////////////////////////////////////////////////////////////////
// DynFloat Class
//

DynFloat::DynFloat()
{
  m_fValue = 0.0;
  m_pcDynExpr = NULL;
}

DynFloat::~DynFloat()
{
  m_fValue = 0.0;
  m_pcDynExpr = NULL;
}

VOID DynFloat::Update()
{
  DynEval(&m_fValue, &m_pcDynExpr, 1);
}

//////////////////////////////////////////////////////////////////////
// DynInt Class
//

DynInt::DynInt()
{
  m_iValue = 0;
  m_pcDynExpr = NULL;
}

DynInt::~DynInt()
{
  m_iValue = 0;
  m_pcDynExpr = NULL;
}

VOID DynInt::Update()
{
  DynEvalInt(&m_iValue, &m_pcDynExpr, 1);
}




