#include "oot.h"
//#include "glsi.h"

// Prototypes of Tcl Callbacks
static int Oot_Tcl_CreateObject(ClientData clientData, Tcl_Interp *interp, 
			    int argc, char **argv);
static int Oot_Tcl_ObjectMethod(ClientData clientData, Tcl_Interp *interp, 
			    int argc, char **argv);
static VOID Oot_Tcl_DestroyObject(ClientData clientData);

// Prototypes of Helper Functions
static VOID Oot_PrintClassMethods(OotClass *pClass);
static VOID Oot_PrintMethod(Tcl_Interp *pInterp, OotMethodEntry *pMethodEntry);
static VOID Oot_PrintClassFields(OotClass *pClass);
static VOID Oot_PrintClassObjects(OotClass *pClass);

//////////////////////////////////////////////////////////////////////////////
// API Functions
//

int Oot_Init(Tcl_Interp * /*pInterp*/)
{
  return TCL_OK;
}

OotClass *Oot_CreateClass(Tcl_Interp *pInterp, const CHAR *pcName, 
			  Oot_Constructor *fpConstructor, 
			  OotMethodEntry *pMethods, UINT_32 uiMethods,
			  const CHAR **cppcFields, UINT_32 uiFields)
{
  UINT_32 i;
  OotMethodEntry *pMethodEntry;

  OotClass *pClass;

  ASSERT( pInterp != NULL && pcName != NULL && fpConstructor != NULL );
  ASSERT( pMethods != NULL && uiMethods > 0 );
  ASSERT( cppcFields != NULL && uiFields > 0 );

  pClass = new OotClass;
  ASSERT( pClass != NULL );

  pClass->pcName = new char[strlen(pcName)+1];
  strcpy(pClass->pcName, pcName);
  pClass->pInterp = pInterp;
  pClass->fpConstructor = fpConstructor;
  pClass->pCHashObjects = new CHashString;
  ASSERT( pClass->pCHashObjects != NULL );
  pClass->pCHashObjects->Init(7);
  pClass->pCHashMethods = new CHashString;
  ASSERT( pClass->pCHashMethods != NULL );
  pClass->pCHashMethods->Init(uiMethods);
  pClass->pMethodEntries = pMethods;
  pClass->uiMethodEntries = uiMethods;
  pClass->cppcFields = cppcFields;
  pClass->uiFields = uiFields;

  for ( i = 0; i < pClass->uiMethodEntries; i++ )
    {
      pMethodEntry = (OotMethodEntry*)(pClass->pCHashMethods->
	               Insert(&(pClass->pMethodEntries[i]), 
			      pClass->pMethodEntries[i].pcName)); 
      if ( pMethodEntry != &(pClass->pMethodEntries[i]) )
	{
	  return NULL;
	}
    }

  Tcl_CreateCommand(pInterp, pClass->pcName, Oot_Tcl_CreateObject, 
		    pClass, Oot_Tcl_DestroyObject);

  return pClass;
}

OotObject *Oot_CreateObject(Tcl_Interp *pInterp, OotClass *pClass, 
			    const CHAR *pcName, ClientData pvClientData)
{
  OotObject *pObject;
  OotObject *pObjectTmp;

  ASSERT( pClass != NULL && pcName != NULL );

  // Check if object already exists
  pObject = (OotObject*)(pClass->pCHashObjects->Peek(pcName));
  ASSERT( pObject == NULL );

  pObject = new OotObject;
  ASSERT( pObject != NULL );
  pObject->pcName = new CHAR[strlen(pcName)+1];
  strcpy(pObject->pcName, pcName);
  pObject->pClass = pClass;
  pObject->pvClientData = pvClientData;

  pObjectTmp = (OotObject*)(pClass->pCHashObjects->Insert(pObject, pcName));
  ASSERT( pObjectTmp == pObject );

  Tcl_CreateCommand(pInterp, pObject->pcName, Oot_Tcl_ObjectMethod, 
		    pObject, NULL);
  
  return pObject;
}

OotObject *Oot_GetObject(OotClass *pClass, const CHAR *pcName)
{
  ASSERT( pClass != NULL && pcName != NULL );

  //printf("Looking up object: %s\n", pcName);
  //pClass->pCHashObjects->Display();
  return (OotObject*)(pClass->pCHashObjects->Peek(pcName));
}

static CHAR gpcOotBufferVar[200];
static CHAR gpcOotBufferValue[200];

VOID Oot_SetObjectField(OotObject *pObject, const CHAR *pcField, BOOL b)
{
  ASSERT( pObject != NULL && pcField != NULL );

  sprintf(gpcOotBufferVar, "%s_%s", pObject->pcName, pcField);
  sprintf(gpcOotBufferValue, "%d", b);
  Tcl_SetVar(pObject->pClass->pInterp, gpcOotBufferVar, gpcOotBufferValue, 
	     TCL_GLOBAL_ONLY);
}

VOID Oot_SetObjectField(OotObject *pObject, const CHAR *pcField, INT_32 i)
{
  ASSERT( pObject != NULL && pcField != NULL );

  sprintf(gpcOotBufferVar, "%s_%s", pObject->pcName, pcField);
  sprintf(gpcOotBufferValue, "%ld", i);
  Tcl_SetVar(pObject->pClass->pInterp, gpcOotBufferVar, gpcOotBufferValue, 
	     TCL_GLOBAL_ONLY);
}

VOID Oot_SetObjectField(OotObject *pObject, const CHAR *pcField, FLOAT f)
{
  ASSERT( pObject != NULL && pcField != NULL );

  sprintf(gpcOotBufferVar, "%s_%s", pObject->pcName, pcField);
  sprintf(gpcOotBufferValue, "%f", f);
  Tcl_SetVar(pObject->pClass->pInterp, gpcOotBufferVar, gpcOotBufferValue, 
	     TCL_GLOBAL_ONLY);
}

VOID Oot_SetObjectField(OotObject *pObject, const CHAR *pcField, DOUBLE d)
{
  ASSERT( pObject != NULL && pcField != NULL );

  sprintf(gpcOotBufferVar, "%s_%s", pObject->pcName, pcField);
  sprintf(gpcOotBufferValue, "%f", d);
  Tcl_SetVar(pObject->pClass->pInterp, gpcOotBufferVar, gpcOotBufferValue, 
	     TCL_GLOBAL_ONLY);
}

VOID Oot_SetObjectField(OotObject *pObject, const CHAR *pcField, const CHAR *pc)
{
  ASSERT( pObject != NULL && pcField != NULL && pc != NULL );

  sprintf(gpcOotBufferVar, "%s_%s", pObject->pcName, pcField);
  strcpy(gpcOotBufferValue, pc);
  Tcl_SetVar(pObject->pClass->pInterp, gpcOotBufferVar, gpcOotBufferValue, 
	     TCL_GLOBAL_ONLY);
}

//////////////////////////////////////////////////////////////////////////////
// Tcl Callbacks
//

// Oot_CreateObject
//   called first from Tcl when an object is constructed
static int Oot_Tcl_CreateObject(ClientData pvClientData, Tcl_Interp *pInterp, 
				int argc, char **argv)
{
  OotClass *pClass;

  ASSERT( pvClientData != NULL && pInterp != NULL );
  ASSERT( argc > 0 && argv != NULL );

  pClass = (OotClass*)pvClientData;
  ASSERT( strcmp(pClass->pcName, argv[0]) == 0 );

  if ( argc == 1 )
    {
      Oot_PrintClassMethods(pClass);
      return TCL_OK;
    }

  if ( strncmp(argv[1], "fields", 6) == 0 )
    {
      Oot_PrintClassFields(pClass);
    } 
  else if ( strncmp(argv[1], "methods", 7) == 0 )
    {
      Oot_PrintClassMethods(pClass);
    } 
  else if ( strncmp(argv[1], "objects", 7) == 0 )
    {
      Oot_PrintClassObjects(pClass);
    } 
  else 
    {
      if ( pClass->fpConstructor(pInterp, pClass, argc-2, &(argv[2])) != TCL_OK )
	{
	  return TCL_ERROR;
	}
    }

  return TCL_OK;
}
/*
static int Oot_Tcl_CreateObject(ClientData pvClientData, Tcl_Interp *pInterp, 
				int argc, char **argv)
{
  OotClass *pClass;
  OotObject *pObject;

  ASSERT( pvClientData != NULL && pInterp != NULL );
  ASSERT( argc > 0 && argv != NULL );

  pClass = (OotClass*)pvClientData;
  ASSERT( strcmp(pClass->pcName, argv[0]) == 0 );

  if ( argc == 1 )
    {
      Oot_PrintClassMethods(pClass);
      return TCL_OK;
    }

  pObject = new OotObject;
  ASSERT( pObject != NULL );
  pObject->pcName = new CHAR[strlen(argv[1])+1];
  strcpy(pObject->pcName, argv[1]);
  pObject->pClass = pClass;

  if ( pClass->fpConstructor(pInterp, pObject, argc-2, &(argv[2])) != TCL_OK )
    {
      return TCL_ERROR;
    }

  Tcl_CreateCommand(pInterp, pObject->pcName, Oot_Tcl_ObjectMethod, 
		    pObject, Oot_DestroyObject);
  
  return TCL_OK;
}
*/
static int Oot_Tcl_ObjectMethod(ClientData pvClientData, Tcl_Interp *pInterp, 
				int argc, char **argv)
{
  OotClass *pClass;
  OotObject *pObject;
  OotMethodEntry *pMethodEntry;
  int iResult;
  int piArgs[4];
  double pdArgs[4];

  ASSERT( pvClientData != NULL && pInterp != NULL );
  ASSERT( argc > 0 && argv != NULL );

  pObject = (OotObject*)pvClientData;
  ASSERT( strcmp(pObject->pcName, argv[0]) == 0 );
  pClass = pObject->pClass;

  if ( argc < 2 )
    {
      Oot_PrintClassMethods(pClass);
      return TCL_OK;
    }

  pMethodEntry = (OotMethodEntry*)(pClass->pCHashMethods->Peek(argv[1]));
  if ( pMethodEntry == NULL )
    {
      return TCL_ERROR;
    }

  iResult = TCL_ERROR;
  switch ( pMethodEntry->uiParamType )
    {
    case OOT_PARAM_TYPE_VOID:
      if ( argc == 2 )
	{
	  iResult = ((Oot_Method_Void*)(pMethodEntry->fpMethod))(pInterp, pObject);
	}
      break;
    case OOT_PARAM_TYPE_BOOL:
      if ( argc == 3 )
	{
	  int i;
	  if ( Tcl_GetBoolean(pInterp, argv[2], &i) == TCL_OK )
	    {
	      iResult = ((Oot_Method_Bool*)(pMethodEntry->fpMethod))(pInterp, pObject, i);
	    }
	}
      break;
    case OOT_PARAM_TYPE_INT:
      int i;
      if ( argc == 3 )
	{
	  if ( Tcl_GetInt(pInterp, argv[2], &i) == TCL_OK )
	    {
	      iResult = ((Oot_Method_Int*)(pMethodEntry->fpMethod))(pInterp, pObject, i);
	    }
	}
      break;
    case OOT_PARAM_TYPE_INT_2:
      if ( argc == 4 )
	{
	  if ( Tcl_GetInt(pInterp, argv[2], &piArgs[0]) == TCL_OK &&
	       Tcl_GetInt(pInterp, argv[3], &piArgs[1]) == TCL_OK )
	    {
	      iResult = ((Oot_Method_Int_2*)(pMethodEntry->fpMethod))(pInterp, pObject, 
								      piArgs[0], piArgs[1]);
	    }
	}
      break;
    case OOT_PARAM_TYPE_INT_3:
      if ( argc == 5 )
	{
	  if ( Tcl_GetInt(pInterp, argv[2], &piArgs[0]) == TCL_OK &&
	       Tcl_GetInt(pInterp, argv[3], &piArgs[1]) == TCL_OK &&
	       Tcl_GetInt(pInterp, argv[4], &piArgs[2]) == TCL_OK )
	    {
	      iResult = ((Oot_Method_Int_3*)(pMethodEntry->fpMethod))(pInterp, pObject, 
								      piArgs[0], piArgs[1],
								      piArgs[2]);
	    }
	}
      break;
    case OOT_PARAM_TYPE_INT_4:
      if ( argc == 6 )
	{
	  if ( Tcl_GetInt(pInterp, argv[2], &piArgs[0]) == TCL_OK &&
	       Tcl_GetInt(pInterp, argv[3], &piArgs[1]) == TCL_OK &&
	       Tcl_GetInt(pInterp, argv[4], &piArgs[2]) == TCL_OK &&
	       Tcl_GetInt(pInterp, argv[5], &piArgs[3]) == TCL_OK )
	    {
	      iResult = ((Oot_Method_Int_4*)(pMethodEntry->fpMethod))(pInterp, pObject, 
								      piArgs[0], piArgs[1],
								      piArgs[2], piArgs[3]);
	    }
	}
      break;
    case OOT_PARAM_TYPE_KEYBOARD:
      if ( argc == 6 )
	{
	  if ( Tcl_GetInt(pInterp, argv[3], &piArgs[1]) == TCL_OK &&
	       Tcl_GetInt(pInterp, argv[4], &piArgs[2]) == TCL_OK &&
	       Tcl_GetInt(pInterp, argv[5], &piArgs[3]) == TCL_OK )
	    {
	      iResult = ((Oot_Method_Keyboard*)(pMethodEntry->fpMethod))(pInterp, pObject, 
								      argv[2], piArgs[1],
								      piArgs[2], piArgs[3]);
	    }
	}
      break;
    case OOT_PARAM_TYPE_FLOAT:
      if ( argc == 3 )
	{
	  double d;
	  if ( Tcl_GetDouble(pInterp, argv[2], &d) == TCL_OK )
	    {
	      iResult = ((Oot_Method_Float*)(pMethodEntry->fpMethod))(pInterp, pObject, (float)d);
	    }
	}
      break;
    case OOT_PARAM_TYPE_FLOAT_3:
      if ( argc == 5 )
	{
	  if ( Tcl_GetDouble(pInterp, argv[2], &pdArgs[0]) == TCL_OK &&
	       Tcl_GetDouble(pInterp, argv[3], &pdArgs[1]) == TCL_OK &&
	       Tcl_GetDouble(pInterp, argv[4], &pdArgs[2]) == TCL_OK )
	    {
	      iResult = ((Oot_Method_Float_3*)(pMethodEntry->fpMethod))(pInterp, pObject, 
					       (float)(pdArgs[0]), (float)(pdArgs[1]),
					       (float)(pdArgs[2]));
	    }
	}
      break;
    case OOT_PARAM_TYPE_FLOAT_4:
      if ( argc == 6 )
	{
	  if ( Tcl_GetDouble(pInterp, argv[2], &pdArgs[0]) == TCL_OK &&
	       Tcl_GetDouble(pInterp, argv[3], &pdArgs[1]) == TCL_OK &&
	       Tcl_GetDouble(pInterp, argv[4], &pdArgs[2]) == TCL_OK &&
	       Tcl_GetDouble(pInterp, argv[5], &pdArgs[3]) == TCL_OK )
	    {
	      iResult = ((Oot_Method_Float_4*)(pMethodEntry->fpMethod))(pInterp, pObject, 
					       (float)(pdArgs[0]), (float)(pdArgs[1]),
					       (float)(pdArgs[2]), (float)(pdArgs[3]));
	    }
	}
      break;
    case OOT_PARAM_TYPE_DOUBLE:
      if ( argc == 3 )
	{
	  double d;
	  if ( Tcl_GetDouble(pInterp, argv[2], &d) == TCL_OK )
	    {
	      iResult = ((Oot_Method_Double*)(pMethodEntry->fpMethod))(pInterp, pObject, d);
	    }
	}
      break;

    case OOT_PARAM_TYPE_PCHAR:
      if ( argc == 3 )
	{
	  iResult = ((Oot_Method_pChar*)(pMethodEntry->fpMethod))(pInterp, pObject, argv[2]);
	}
      break;

    default:
      iResult = ((Oot_Method_Other*)(pMethodEntry->fpMethod))(pInterp, pObject, argc-2, &(argv[2]));
      break;
    }

  if ( iResult == TCL_ERROR )
    {
      Oot_PrintMethod(pInterp, pMethodEntry);
    }

  return iResult; 
}

// Oot_Destroy 
//   called first from Tcl when an object is constructed
static VOID Oot_Tcl_DestroyObject(ClientData /*pvClientData*/)
{
  //OotClass *pClass;

  //pClass = (OotClass*)pvClientData;
}

//////////////////////////////////////////////////////////////////////////////
// These are not called from Tcl
//

static VOID Oot_PrintClassFields(OotClass *pClass)
{
  UINT_32 i;
  Tcl_Interp *pInterp;

  ASSERT( pClass != NULL );

  pInterp = pClass->pInterp;

  Tcl_AppendResult(pInterp, "class ", pClass->pcName, ":", NULL); 
  for ( i = 0; i < pClass->uiFields; i++ )
    {
      Tcl_AppendResult(pInterp, "\n  ", pClass->cppcFields[i], NULL); 
    }
}

static VOID Oot_PrintClassMethods(OotClass *pClass)
{
  UINT_32 i;
  Tcl_Interp *pInterp;

  ASSERT( pClass != NULL );

  pInterp = pClass->pInterp;

  Tcl_AppendResult(pInterp, "class ", pClass->pcName, ":", NULL); 
  for ( i = 0; i < pClass->uiMethodEntries; i++ )
    {
      Tcl_AppendResult(pInterp, "\n  ", NULL); 
      Oot_PrintMethod(pInterp, &(pClass->pMethodEntries[i])); 
    }
}

static const CHAR *cppcParamType[OOT_PARAM_TYPE_MAX] = 
{
  "",
  " <bool>",
  " <int>",
  " <float>",
  " <double>",
  " ?args?",
  " <int> <int>",
  " <int> <int> <int>",
  " <int> <int> <int> <int>",
  " <float> <float> <float>",
  " <float> <float> <float> <float>",
  " <keycode> <int> <int> <int>"
};

static VOID Oot_PrintMethod(Tcl_Interp *pInterp, OotMethodEntry *pMethodEntry)
{
  ASSERT( pInterp != NULL && pMethodEntry != NULL );
  ASSERT( pMethodEntry->uiParamType < OOT_PARAM_TYPE_MAX );

  Tcl_AppendResult(pInterp, pMethodEntry->pcName, 
		   cppcParamType[pMethodEntry->uiParamType], NULL); 
}

static VOID Oot_PrintClassObjects(OotClass *pClass)
{
  ASSERT( pClass != NULL );

  pClass->pCHashObjects->Display();
}

