#include "point.h"

#include "vector.h"
#include "dynamics.h"

Point::Point()
{
  Null();
}

Point::Point(FLOAT x, FLOAT y, FLOAT z, FLOAT w/* = 1.0f*/)
{
  m_pt[X] = x;
  m_pt[Y] = y;
  m_pt[Z] = z;
  m_pt[W] = w;
}

Point::~Point()
{
}

VOID Point::Null()
{
  m_pt[X] = 0.0f;
  m_pt[Y] = 0.0f;
  m_pt[Z] = 0.0f;
  m_pt[W] = 1.0f;
}

BOOL Point::Homogenize()
{
  if ( m_pt[W] == 0.0f )
    {
      return FALSE;
    }

  m_pt[X] = m_pt[X] / m_pt[W];
  m_pt[Y] = m_pt[Y] / m_pt[W];
  m_pt[Z] = m_pt[Z] / m_pt[W];
  m_pt[W] = 1.0f;

  return TRUE;
}

Point &Point::operator=(const Point &pt)
{
  m_pt[0] = pt(0);
  m_pt[1] = pt(1);
  m_pt[2] = pt(2);
  m_pt[3] = pt(3);

  return *this;
}

Point &Point::operator=(const Vector &v)
{
  m_pt[0] = v(0);
  m_pt[1] = v(1);
  m_pt[2] = v(2);
  m_pt[3] = (FLOAT)(v(3) + 1.0);

  return *this;
}

Point &Point::operator+=(const Vector &v)
{
  m_pt[0] = m_pt[0] + v(0);
  m_pt[1] = m_pt[1] + v(1);
  m_pt[2] = m_pt[2] + v(2);
  m_pt[3] = m_pt[3] + v(3);

  return *this;
}

Point &Point::operator-=(const Vector &v)
{
  m_pt[0] = m_pt[0] - v(0);
  m_pt[1] = m_pt[1] - v(1);
  m_pt[2] = m_pt[2] - v(2);
  m_pt[3] = m_pt[3] - v(3);

  return *this;
}

Point &Point::operator+=(const Point &pt)
{
  m_pt[0] = m_pt[0] + pt(0);
  m_pt[1] = m_pt[1] + pt(1);
  m_pt[2] = m_pt[2] + pt(2);
  m_pt[3] = m_pt[3] + pt(3);

  return *this;
}

Point &Point::operator*=(const FLOAT f)
{
  m_pt[0] = m_pt[0] * f;
  m_pt[1] = m_pt[1] * f;
  m_pt[2] = m_pt[2] * f;
  m_pt[3] = m_pt[3] * f;

  return *this;
}

Point &Point::operator/=(const FLOAT f)
{
  m_pt[0] = m_pt[0] / f;
  m_pt[1] = m_pt[1] / f;
  m_pt[2] = m_pt[2] / f;
  m_pt[3] = m_pt[3] / f;

  return *this;
}

BOOL Point::operator==(const Point &pt) const
{
  FLOAT fDot, fDotV, fVDotV;

  fDot = this->Dot(*this);
  fDotV = this->Dot(pt);
  fVDotV = pt.Dot(pt);

  return ((fDot > (fDotV - EPSILON)) && (fDot < (fDotV + EPSILON))) &&
         ((fDot > (fVDotV - EPSILON)) && (fDot < (fVDotV + EPSILON)));
}

BOOL Point::operator!=(const Point &pt) const
{
  return !(*this == pt);
}

FLOAT Point::Dot(const Vector &v) const
{
  return (((*this)(0)*v(0)) + ((*this)(1)*v(1)) + 
          ((*this)(2)*v(2)) + ((*this)(3)*v(3))); 
}

BOOL Point::operator<(const Point &pt) const
{
  return ( (*this)(X) < pt(X) &&
	   (*this)(Y) < pt(Y) &&
	   (*this)(Z) < pt(Z) );
}

BOOL Point::operator<=(const Point &pt) const
{
  return ( (*this)(X) <= pt(X) &&
	   (*this)(Y) <= pt(Y) &&
	   (*this)(Z) <= pt(Z) );
}

BOOL Point::operator>(const Point &pt) const
{
  return ( (*this)(X) > pt(X) &&
	   (*this)(Y) > pt(Y) &&
	   (*this)(Z) > pt(Z) );
}

BOOL Point::operator>=(const Point &pt) const
{
  return ( (*this)(X) >= pt(X) &&
	   (*this)(Y) >= pt(Y) &&
	   (*this)(Z) >= pt(Z) );
}

FLOAT Point::Dot(const Point &pt) const
{
  return (((*this)(0)*pt(0)) + ((*this)(1)*pt(1)) + 
          ((*this)(2)*pt(2)) + ((*this)(3)*pt(3))); 
}

VOID Point::Display() const
{
  UINT_32 i;

  printf("< ");
  for ( i = 0; i < 4; i++)
    {
      printf("%3.3f ", (*this)(i));
    }
  printf(">\n");
}

//////////////////////////////////////////////////////////////////////
// DynPoint Class
//

DynPoint::DynPoint()
{
  Null();

  m_ppcDynExpr[0] = NULL;
  m_ppcDynExpr[1] = NULL;
  m_ppcDynExpr[2] = NULL;
  m_ppcDynExpr[3] = NULL;
}

DynPoint::DynPoint(FLOAT x, FLOAT y, FLOAT z, FLOAT w/* = 0.0f*/)
{
  m_pt[0] = x;
  m_pt[1] = y;
  m_pt[2] = z;
  m_pt[3] = w;

  m_ppcDynExpr[0] = NULL;
  m_ppcDynExpr[1] = NULL;
  m_ppcDynExpr[2] = NULL;
  m_ppcDynExpr[3] = NULL;
}

DynPoint::~DynPoint()
{
  m_ppcDynExpr[0] = NULL;
  m_ppcDynExpr[1] = NULL;
  m_ppcDynExpr[2] = NULL;
  m_ppcDynExpr[3] = NULL;
}

VOID DynPoint::Update()
{
  DynEval(m_pt, m_ppcDynExpr, 4);
}
