#include "slideworld.h"

static CHAR gpcBuffer[200];
const FLOAT gcfFrameRate = 30.0;
//const FLOAT gcfFrameRate = 2.0;
const FLOAT gcfFrameTime = (FLOAT)1.0 / gcfFrameRate;

Tcl_Interp *CSLIDEWorld::m_pInterp = NULL;
CSLIDEWorld *CSLIDEWorld::m_pCSLIDEWorld = NULL;

DOUBLE CSLIDEWorld::m_dSecondsPerFrameRequested = gcfFrameTime;
DOUBLE CSLIDEWorld::m_dUpdateTime = 0;


//////////////////////////////////////////////////////////////////////
// CSLIDEWorld Class
//

CSLIDEWorld::CSLIDEWorld()
{
}

CSLIDEWorld::~CSLIDEWorld()
{
}

VOID CSLIDEWorld::Init()
{
  m_listCWindow.Init();
  m_hashCWindow.Init(1);

  InitDefaults();

  // Dynamics
  m_bDynamicsPaused = FALSE;
  m_dSecondsPerFrameRequested = gcfFrameTime;
  m_ttkWakeUpIDDynamics = 0;
  m_dTimeBase = 0.0;
  m_uiFrame = 0;
}

VOID CSLIDEWorld::InitDefaults()
{
}

VOID CSLIDEWorld::Uninit()
{
  CListIterInt iter;
  CWindow *pCWindow;

  iter.Init(&m_listCWindow);

  for ( pCWindow = (CWindow*)(iter.PeekFirst());
	pCWindow != NULL;
	pCWindow = (CWindow*)(iter.PeekNext()) )
    {
      delete pCWindow;
    }
}

BOOL CSLIDEWorld::Initialized()
{
  return TRUE;
}

BOOL CSLIDEWorld::InsertWindow(CWindow *pCWindow)
{
  CHAR *pcID;

  ASSERT( pCWindow != NULL );

  pcID = pCWindow->GetSLIDEID();

  m_hashCWindow.Insert(pCWindow, (UINT_32)pcID);
  m_listCWindow.InsertLast(pCWindow, 0);

  return TRUE;
}

CWindow *CSLIDEWorld::PeekWindow(CHAR *pcID)
{
  ASSERT( pcID != NULL );

  return (CWindow*)(m_hashCWindow.Peek((UINT_32)pcID));
}

BOOL CSLIDEWorld::CheckSemantics()
{
  BOOL b, bCorrect;
  CListIterInt iter;
  CWindow *pCWindow;

  bCorrect = TRUE;

  iter.Init(&m_listCWindow);

  for ( pCWindow = (CWindow*)(iter.PeekFirst());
	pCWindow != NULL;
	pCWindow = (CWindow*)(iter.PeekNext()) )
    {
      b = pCWindow->CheckSemantics(this);
      bCorrect = bCorrect && b;
    }

  return bCorrect;
}

BOOL CSLIDEWorld::Valid()
{
  return TRUE;
}

VOID CSLIDEWorld::Preprocess()
{
  CListIterInt iter;
  CWindow *pCWindow;

  // Preprocessing the Windows will actually create the windows and
  // bring them up on the screen.  This is why we do this last.
  iter.Init(&m_listCWindow);

  for ( pCWindow = (CWindow*)(iter.PeekFirst());
	pCWindow != NULL;
	pCWindow = (CWindow*)(iter.PeekNext()) )
    {
      pCWindow->Preprocess();
    }
}

VOID CSLIDEWorld::WriteSLIDE(FILE *fpOutputFile)
{
  CListIterInt iter;
  CWindow *pCWindow;

  ASSERT( fpOutputFile != NULL );

  iter.Init(&m_listCWindow);

  for ( pCWindow = (CWindow*)(iter.PeekFirst());
	pCWindow != NULL;
	pCWindow = (CWindow*)(iter.PeekNext()) )
    {
      pCWindow->WriteSLIDE(fpOutputFile);
    }
}

UINT_32 CSLIDEWorld::GetErrorCount()
{
  return CSLIDEPrimitive::GetErrorCount();
}

UINT_32 CSLIDEWorld::GetWarningCount()
{
  return CSLIDEPrimitive::GetWarningCount();
}




